/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture.NEW_LINK;

import java.util.List;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewer;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IClickController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.MVCBundleTag;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.MVCBundlePartBase;

import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.control.MenuItem;
import javafx.scene.input.Dragboard;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseEvent;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;

/** Base class for {@link IController} implementations with move and link operation. */
public abstract class ControllerBase extends MVCBundlePartBase implements IController {
	/** Constructor. */
	public ControllerBase(IMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnShiftMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		return new ClickControllerBase() {
			/** {@inheritDoc} */
			@Override
			public Change singleClick(MouseEvent event, Node node,
					DiagramCoordinate diagramLocation) {
				if(!selectOnClick()) {
					return null;
				}
				if(event.isShiftDown()) {
					getViewer().handleShiftSelectionOf(getMVCBundle());
				} else {
					getViewer().setSingleSelectedMVCBundle(getMVCBundle());
				}
				return null;
			}

			/** {@inheritDoc} */
			@Override
			public Change secondaryClick(MouseEvent event, Node node,
					DiagramCoordinate diagramLocation) {
				DiagramViewer viewer = getViewer();
				if(selectOnClick()) {
					viewer.setSingleSelectedMVCBundle(getMVCBundle());
				}
				if(event.isControlDown()) {
					return () -> delete();
				}
				viewer.showContextMenu(node, diagramLocation, getMVCBundle());
				return null;
			}
		};
	}

	/** Returns whether the visual should be selected on single click. */
	protected boolean selectOnClick() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean handleExternalDNDDragOver(Dragboard db, DiagramCoordinate diagramLocation) {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean handleExternalDNDDrop(Dragboard db, DiagramCoordinate diagramLocation) {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(allowLink() && gesture == NEW_LINK) {
			return getLinkCreationController();
		}
		return null;
	}

	/** Returns whether this visual can be linked. */
	protected boolean allowLink() {
		return true;
	}

	/** Returns the default {@link IDragController} for creating links. */
	protected IDragController getLinkCreationController() {
		return defaultLinkController;
	}

	/** The default {@link IDragController} for link creation. */
	private final IDragController defaultLinkController = new DragControllerBase() {
		@Override
		public void safeDragStarted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			diagramLocation = getVisual().getLinkAnchorage(diagramLocation);
			getViewer().startNewLinkLineFeedback(bundle, node, diagramLocation);
		}

		@Override
		protected void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			diagramLocation = bundle.getVisual().getLinkAnchorage(diagramLocation);
			getViewer().updateNewLinkLineFeedback(diagramLocation,
					newLinkFeedback(getMVCBundle(), startDiagramLocation, bundle, diagramLocation));
		}

		@Override
		protected Change safeDragCompleted(IMVCBundle endBundle, Node node,
				DiagramCoordinate endDiagramLocation) {
			IMVCBundle startBundle = getMVCBundle();
			getViewer().terminateNewLinkLineFeedback();
			if(canLink(startBundle, startDiagramLocation, endBundle, endDiagramLocation)) {
				// Need to copy the location to a local variable to make sure that it is passed
				// by value to the returned closure. Otherwise the closure will capture 'this' and
				// link will see a future value of startDiagramLocation, likely null because
				// startDiagramLocation is being set to null right after this return.
				final DiagramCoordinate startDiagramLocation_ = startDiagramLocation;
				return () -> link(startBundle, startDiagramLocation_, endBundle,
						endDiagramLocation);
			}
			return null;
		}
	};

	/**
	 * Returns whether a new link can be created between the given bundles and locations.
	 * 
	 * @param startBundle
	 *            the {@link IMVCBundle} of the start node of the link
	 * @param startLocation
	 *            the location of the start of the link
	 * @param endBundle
	 *            the {@link IMVCBundle} of the end node of the link
	 * @param endLocation
	 *            the location of the end of the link
	 * @return if the link can be created; default is true
	 */
	protected boolean canLink(IMVCBundle startBundle, DiagramCoordinate startLocation,
			IMVCBundle endBundle, DiagramCoordinate endLocation) {
		return true;
	}

	/**
	 * Creates a link between the bundles at the given, respective locations.
	 * 
	 * @param startBundle
	 *            the {@link IMVCBundle} of the start node of the link
	 * @param startLocation
	 *            the location of the start of the link
	 * @param endBundle
	 *            the {@link IMVCBundle} of the end node of the link
	 * @param endLocation
	 *            the location of the end of the link
	 */
	protected void link(IMVCBundle startBundle, DiagramCoordinate startLocation,
			IMVCBundle endBundle, DiagramCoordinate endLocation) {
		// the default does nothing
	}

	/**
	 * Determines the feedback color of for creating new links between the given bundles.
	 * 
	 * @param startBundle
	 *            the {@link IMVCBundle} of the start node of the link
	 * @param startLocation
	 *            the location of the start of the link
	 * @param endBundle
	 *            the {@link IMVCBundle} of the end node of the link
	 * @param endLocation
	 *            the location of the end of the link
	 * @return the feedback line color
	 */
	protected Paint newLinkFeedback(IMVCBundle startBundle, DiagramCoordinate startLocation,
			IMVCBundle endBundle, DiagramCoordinate endLocation) {
		return Color.BLACK;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isPossibleLinkTarget() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForNewLinkFrom(IMVCBundle linkStartBundle) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForMoveLink(ILinkMVCBundle linkBundle,
			IAnchorageMVCBundle movedAnchorageBundle) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void delete() {
		// the default does nothing
	}

	/**
	 * The default implementation for key press controller allowing move with cursor keys and delete
	 * with the delete key.
	 */
	private final IKeyPressController defaultKeyPressController = new IKeyPressController() {
		@Override
		public Change keyEvent(KeyEvent event, IMVCBundle bundle, Node source,
				DiagramCoordinate mouseLocation) {
			DiagramViewer viewer = getViewer();
			KeyCode code = event.getCode();
			if(code == KeyCode.DELETE) {
				return new Change() {
					@Override
					public void applyChange() {
						delete();
					}
				};
			} else if(code == KeyCode.A && event.isControlDown()) {
				viewer.selectAll();
			}
			return null;
		}
	};

	/** {@inheritDoc} */
	@Override
	public IKeyPressController getKeyPressController(Node node,
			DiagramCoordinate lastMouseLocation) {
		return defaultKeyPressController;
	}

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		switch(edg) {
			case NEW_LINK:
				return allowLink() ? Cursor.CROSSHAIR : Cursor.DEFAULT;
			default:
				return Cursor.DEFAULT;
		}
	}

	/** {@inheritDoc} */
	@Override
	public IModelChangeProvider getModelChangeProvider() {
		return null;
	}
}
