/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange.locationFeedbackChange;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.DraggingUtils.clampAndSnapMoveLocationToGridCenterPoint;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual.END_OF_LINK_BEND_POINT_INDEX;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual.NO_BEND_POINT_INDEX;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual.SPECIAL_BEND_POINT_INDEX_CAP;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual.START_OF_LINK_BEND_POINT_INDEX;

import java.util.List;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewer;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerSelection;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IClickController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.MVCBundleTag;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.MVCBundlePartBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual;

import javafx.geometry.Point2D;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.control.MenuItem;
import javafx.scene.input.Dragboard;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseEvent;

/**
 * Base class implementing {@link IController} for links. It provides dragging implementation for
 * visuals of bend-points and of the start or end of the link. It also handles creation and removal
 * of bend-points (including keyboard interaction).
 * <P>
 * Sub-classes must implement the abstract methods, which are called once the respective operation
 * is completed, and write the results back into the underlying data model.
 */
public abstract class LinkControllerBase extends MVCBundlePartBase implements IController {
	/** Stores the index of the currently moving bend-point. */
	private int currentBendPointIndex = 0;

	/** Constructor. */
	public LinkControllerBase(ILinkMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public IModelChangeProvider getModelChangeProvider() {
		return null;
	}

	/** Create a new bend point with the given mouse click coordinates. */
	protected abstract void createBendPointAt(int bpIndex, DiagramCoordinate location);

	/**
	 * Moves the model element.
	 * 
	 * @param bpIndex
	 *            the index of moved bend-point
	 * @param dx
	 *            the x delta
	 * @param dy
	 *            the y delta
	 */
	protected abstract void updateModelAfterBendPointMove(int bpIndex, double dx, double dy);

	/** Deletes the current bend-point. */
	protected abstract void deleteBendPoint(int bpIndex);

	/**
	 * Moves the given end of the link to the new anchorage. The given bend-point
	 */
	protected abstract void updateModelAfterLinkMove(ILinkMVCBundle linkBundle,
			IMVCBundle startBundle, IMVCBundle endBundle);

	/** Removes the given bend point. */
	protected abstract void deleteLink();

	/** Sets the current bend-point index. */
	protected final void setCurrentBendPointIndex(int bendPointIndex) {
		currentBendPointIndex = bendPointIndex;
	}

	/** {@inheritDoc} */
	@Override
	public ILinkMVCBundle getMVCBundle() {
		// wild cast works: see constructor
		return (ILinkMVCBundle)super.getMVCBundle();
	}

	/** Returns the link visual. */
	protected ILinkVisual getLinkVisual() {
		return getMVCBundle().getVisual();
	}

	/** Returns the {@link IClickController} for creating bend-points and selecting the link. */
	protected IClickController getBendPointCreateAndLinkSelectController() {
		return new ClickControllerBase() {
			@Override
			public Change singleClick(MouseEvent event, Node node,
					DiagramCoordinate diagramLocation) {
				DiagramViewer viewer = getViewer();
				if(event.isControlDown()) {
					// link must be primary selected
					if(!viewer.isPrimarySelected(getMVCBundle())) {
						return null;
					}
					// control+click on visible or selection line creates a bend point
					DiagramViewerFeatures features = viewer.getFeatures();
					DiagramCoordinate snappedLocation = features.snapToCenter(diagramLocation);
					return () -> createBendPointAt(currentBendPointIndex, snappedLocation);
				}
				if(event.isShiftDown()) {
					viewer.handleShiftSelectionOf(getMVCBundle());
				} else {
					viewer.setSingleSelectedMVCBundle(getMVCBundle());
				}
				return null;
			}

			@Override
			public Change secondaryClick(MouseEvent event, Node node,
					DiagramCoordinate locationOnNode) {
				if(event.isControlDown()) {
					// link must be primary selected
					DiagramViewer viewer = getViewer();
					if(viewer.isPrimarySelected(getMVCBundle())) {
						// control+right-click deletes the link
						return () -> deleteLink();
					}
				}
				return null;
			}
		};
	}

	/** Returns the {@link IDragController} for moving bend-points. */
	protected IDragController getBendPointMoveController() {
		return new DragControllerBase() {
			@Override
			public void safeDragInProgress(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				Point2D delta = diagramLocation.subtract(startDiagramLocation);
				moveBendPointVisually(currentBendPointIndex, delta);
				getViewer().updateVisual(getMVCBundle());
			}

			@Override
			public Change safeDragCompleted(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				Point2D delta = diagramLocation.subtract(startDiagramLocation);
				return () -> moveBendPointInModel(currentBendPointIndex, delta);
			}
		};
	}

	/**
	 * Updates the bend-point visual(s) during the dragging phase of the move operation. Sub-classes
	 * may override to move other bend-points than only the current bend-point.
	 */
	protected void moveBendPointVisually(int bpIndex, Point2D delta) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, delta.getX(), delta.getY());
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, chg);
	}

	/**
	 * Updates the bend-point visual(s) during the dragging phase of the move operation. Sub-classes
	 * may override to move other bend-points than only the current bend-point.
	 */
	protected void moveBendPointVisually(int bpIndex, FeedbackChange change) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, change.getDeltaX(), change.getDeltaY());
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, chg);
	}

	/** Returns the clamped and snapped move delta feedback change for the given bend point. */
	protected final FeedbackChange getMoveDeltaFeedback(int bpIndex, Point2D delta) {
		return getMoveDeltaFeedback(bpIndex, delta.getX(), delta.getY());
	}

	/** Returns the clamped and snapped move delta feedback change for the given bend point. */
	protected final FeedbackChange getMoveDeltaFeedback(int bpIndex, double dx, double dy) {
		DiagramCoordinate bpModelLocation = getLinkVisual().getBendPointLocation(bpIndex);
		DiagramViewerFeatures features = getViewer().getFeatures();
		if(clampAndSnapBendPointToGridCenter()) {
			return clampAndSnapMoveLocationToGridCenterPoint(bpModelLocation.getX(),
					bpModelLocation.getY(), dx, dy, features);
		}
		if(clampAndSnapBendPointToGrid()) {
			return DraggingUtils.clampAndSnapMoveLocationToGridPoint(bpModelLocation.getX(),
					bpModelLocation.getY(), dx, dy, features);
		}
		return locationFeedbackChange(dx, dy);
	}

	/**
	 * Returns whether bend point positions should be snapped to the center of four grid points.
	 * This mode is the default and has priority over {@link #clampAndSnapBendPointToGrid()}.
	 */
	protected boolean clampAndSnapBendPointToGridCenter() {
		return true;
	}

	/** Returns whether bend point positions should be snapped to grid points. */
	protected boolean clampAndSnapBendPointToGrid() {
		return false;
	}

	/**
	 * Updates the bend-point model element(s) after the dragging of the move operation is complete.
	 */
	protected void moveBendPointInModel(int bpIndex, Point2D delta) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, delta.getX(), delta.getY());
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, null);
		updateModelAfterBendPointMove(bpIndex, chg.getDeltaX(), chg.getDeltaY());
	}

	/**
	 * Updates the bend-point model element(s) after the dragging of the move operation is complete.
	 */
	protected void moveBendPointInModel(int bpIndex, FeedbackChange delta) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, delta.getDeltaX(), delta.getDeltaY());
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, null);
		updateModelAfterBendPointMove(bpIndex, chg.getDeltaX(), chg.getDeltaY());
	}

	/** Returns the {@link IClickController} for deleting bend-points. */
	protected IClickController getBendPointDeleteController() {
		return new ClickControllerBase() {
			@Override
			public Change secondaryClick(MouseEvent event, Node node,
					DiagramCoordinate diagramLocation) {
				if(event.isControlDown()) {
					// bundle must be primary selection
					DiagramViewer viewer = getViewer();
					if(viewer.isPrimarySelected(getMVCBundle())) {
						// control+right-click on selection handle of bend point removes it
						return () -> deleteBendPoint(currentBendPointIndex);
					}
				}
				return null;
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		DiagramViewerSelection selection = getViewer().getSelection();
		ILinkMVCBundle linkBundle = getMVCBundle();
		if(selection.getPrimarySelection() == linkBundle) {
			// Only bend-point handles create drag controllers
			if(!getLinkVisual().isBendPointHandle(node)) {
				return null;
			}
			// links with primary selection show move handles
			ILinkVisual linkVisual = getLinkVisual();
			int bpIndex = linkVisual.getBendPointIndex(node);
			if(bpIndex == START_OF_LINK_BEND_POINT_INDEX ||
					bpIndex == END_OF_LINK_BEND_POINT_INDEX) {
				return getLinkMoveController(bpIndex);
			}
			if(bpIndex == NO_BEND_POINT_INDEX) {
				return null;
			}
			setCurrentBendPointIndex(bpIndex);
			return getBendPointMoveController();
		}
		if(selection.isSelected(linkBundle)) {
			// selected links may be moved by group move operation
			return new DragControllerBase() {
				/** {@inheritDoc} */
				@Override
				protected void safeDragInProgress(IMVCBundle bundle, Node node,
						DiagramCoordinate diagramLocation) {
					FeedbackChange fc = getDragDelta(diagramLocation);
					int numerOfBendPoints = getNumerOfBendPoints();
					for(int i = 0; i < numerOfBendPoints; i++) {
						moveBendPointVisually(i, fc);
					}
				}

				/** {@inheritDoc} */
				@Override
				protected Change safeDragCompleted(IMVCBundle bundle, Node node,
						DiagramCoordinate diagramLocation) {
					FeedbackChange fc = getDragDelta(diagramLocation);
					return () -> {
						int numerOfBendPoints = getNumerOfBendPoints();
						for(int i = 0; i < numerOfBendPoints; i++) {
							moveBendPointInModel(i, fc);
						}
					};
				}
			};
		}
		return null;
	}

	/**
	 * Returns an {@link IDragController} for relocating the end of this link. {@code bpIndex} is
	 * either {@link ILinkVisual#START_OF_LINK_BEND_POINT_INDEX} or
	 * {@link ILinkVisual#END_OF_LINK_BEND_POINT_INDEX} depending on which part of the link is
	 * moved.
	 */
	protected IDragController getLinkMoveController(final int bpIndex) {
		return new DragControllerBase() {
			@Override
			protected void safeDragStarted(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				ILinkMVCBundle linkBundle = getMVCBundle();
				IAnchorageMVCBundle anchorage;
				if(bpIndex == START_OF_LINK_BEND_POINT_INDEX) {
					anchorage = linkBundle.getStartAnchorage();
				} else {
					anchorage = linkBundle.getEndAnchorage();
				}
				getLinkVisual().disableFeedback();
				getViewer().startMoveLinkFeedback(linkBundle, anchorage);
			}

			@Override
			public void safeDragInProgress(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				Point2D linkAnchorageLocation =
						bundle.getVisual().getLinkAnchorage(diagramLocation);
				Point2D delta;
				if(linkAnchorageLocation != null) {
					delta = linkAnchorageLocation.subtract(startDiagramLocation);
				} else {
					delta = diagramLocation.subtract(startDiagramLocation);
				}
				FeedbackChange chg = locationFeedbackChange(delta.getX(), delta.getY());
				getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, chg);
				getViewer().updateVisual(getMVCBundle());
			}

			@Override
			public Change safeDragCompleted(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				getLinkVisual().enableFeedback();
				getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, null);
				DiagramViewer viewer = getViewer();
				viewer.terminateMoveLinkFeedback();
				ILinkMVCBundle linkBundle = getMVCBundle();
				if(bpIndex == START_OF_LINK_BEND_POINT_INDEX) {
					return () -> updateModelAfterLinkMove(linkBundle, bundle,
							linkBundle.getEndAnchorage());
				}
				return () -> updateModelAfterLinkMove(linkBundle, linkBundle.getStartAnchorage(),
						bundle);
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		int bpIndex = getLinkVisual().getBendPointIndex(node);
		if(bpIndex < SPECIAL_BEND_POINT_INDEX_CAP) {
			return null;
		}
		setCurrentBendPointIndex(bpIndex);
		if(getLinkVisual().isBendPointHandle(node)) {
			return getBendPointDeleteController();
		}
		return getBendPointCreateAndLinkSelectController();
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnShiftMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IKeyPressController getKeyPressController(Node node,
			DiagramCoordinate lastMouseLocation) {
		return new IKeyPressController() {
			@Override
			public Change keyEvent(KeyEvent event, IMVCBundle bundle, Node source,
					DiagramCoordinate mouseLocation) {
				KeyCode code = event.getCode();
				if(code == KeyCode.DELETE) {
					return () -> deleteLink();
				} else if(code == KeyCode.A && event.isControlDown()) {
					getViewer().selectAll();
				}
				return null;
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isPossibleLinkTarget() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForNewLinkFrom(IMVCBundle linkStartBundle) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForMoveLink(ILinkMVCBundle linkBundle,
			IAnchorageMVCBundle movedAnchorageBundle) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		switch(edg) {
			case MOVE_BENDPOINT:
				return allowBendPointMove() ? Cursor.MOVE : Cursor.DEFAULT;
			case NEW_BENDPOINT:
				return Cursor.CROSSHAIR;
			case RECONNECT_LINK:
				return allowLinkMove() ? Cursor.MOVE : Cursor.DEFAULT;
			default:
				return Cursor.DEFAULT;
		}
	}

	/** Determines whether bend points can be moved. */
	protected boolean allowBendPointMove() {
		return true;
	}

	/** Determines whether link start and end can be relocated. */
	protected boolean allowLinkMove() {
		return true;
	}

	/** Returns the number of bend-points in the model. */
	protected abstract int getNumerOfBendPoints();

	/** {@inheritDoc} */
	@Override
	public boolean handleExternalDNDDragOver(Dragboard db, DiagramCoordinate diagramLocation) {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean handleExternalDNDDrop(Dragboard db, DiagramCoordinate diagramLocation) {
		return false;
	}
}
